//
// JUMA-TRX2 driver for ISD1700 Voice recorder chip
// software driver for SPI2 interface
//
// Juha Niinikoski, OH2NLT 10.03.2007
//
// 
// MPLAB C30 version
//
// MCU dsPIC30F6014A
//
// Memory banks changed, voice memory option installed flag 17.03.2007
// UART2 added 02.01.2008
// <stdio.h> added,  lcdpbuff[40] type changed to char 05.03.2015

#include <stdio.h>
#include "juma-trx2.h"						// JUMA-TRX2 hardware specific definitions
#include "lcd-trx2.h"						// LCD specific definitions

// external references
// external functions

// EEPROM structures
#include "trx2_eeprom.h"			// get EEPROM structure definitions

extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval)/2];
	}eeprom;

extern int kbhit2(void);					// UART2 input check

extern int kbhit(void);						// UART1 input check
extern unsigned char getch(void);

extern void clear_lcd(void);
extern void set_cur_lcd(unsigned char);
extern void lcd_putst(register const char *);

extern void disp_bar_meter(void);			// measure, calculate & display PWR or S meter

// scaf clock control
extern void stop_scaf_clock(void);			// stop scaf clocking
extern void start_scaf_clock(void);			// start scaf clocking

// adc
extern int convert_adc12(unsigned int adchs);	// ADC

// external variables
extern char lcdpbuff[40];					// LCD print buffer
extern int seconds;							// free running second counter

extern int isd_1700_installed;				// voice memory board found

extern int sq_timer;						// squelch timer

extern int fwd_adval;						// forward & reverse power raw values, measured only when TX

// ISD17240 memory division table for JUMA-TRX2 application
// resolution is 0,125s / lsb bit
const int ISD_MEM [10] [2] = {
{0x10, 0x3FF},					// [0] = recording memory, len = 1008, 126s
{0x400, 0x4AF},					// [1] = MEM 1, len = 88, 11s
{0x4B0, 0x507},					// [2] = MEM 2, len = 88
{0x508, 0x55F},					// [3] = MEM 3, len = 88
{0x560, 0x5B7},					// [4] = MEM 4, len = 88
{0x5B8, 0x60F},					// [5] = MEM 5, len = 88
{0x610 , 0x667},				// [6] = MEM 5, len = 88
{0x668 , 0x6BF},				// [7] = MEM 5, len = 88
{0x6C0 , 0x717},				// [8] = MEM 5, len = 88
{0x718 , 0x78F}					// [9] = MEM 5, len = 119, 14s
};


// ISD17240 Voice Recorder I/O
// desined in trx2.h

//#define ISD_PLAY_BAND _LATG15		//out Play to Band switch, 1=play band

//#define ISD_RESET	_LATG14			//out ISD17240 nRESET
//#define ISD_RDY	_RG12			//in  ISD17240 Ready / INT

//#define ISD_SS	_LATG9			//out ISD17240 SPI nSS
//#define ISD_SCLK	_LATG6			//out ISD17240 SPI SCLK
//#define ISD_MOSI	_LATG8			//out ISD17240 SPI MOSI
//#define ISD_MISO	_RG7			//in  ISD17240 SPI MISO
//#define ISD_17240	0xE0			//ISD17240 Device ID

// ISD1700 commands

// 2-byte commands, second byte =0x00
#define PU			0x01			// Power Up
#define STOP		0x02			// Stop current operation
#define RESET		0x03			// Reset the device
#define CLR_INT		0x04			// Clear int & EOM bit
#define PD			0x07			// Power Down
#define PLAY		0x40			// Play from current location
#define REC			0x41			// Record from current location
#define ERASE		0x42			// Erase current message to EOM
#define G_ERASE		0x43			// Erase all messages
#define WR_NVCFG	0x46			// Write APC to NVCFG
#define LD_NVCFG	0x47			// Load APC from NVCFG
#define FWD			0x48			// Forward playback pointer to next msg
#define CHK_MEM		0x49			// Check circular memory
#define EXTCLK		0x4A			// Enable/disable ext clock mode

// 3-byte commands
#define RD_STATUS	0x05			// Returns current row counter & oper status
#define DEVID		0x09			// Read device ID register
#define WR_APC1		0x45			// Write D11:D0 into the APC reg, VOL from pin
#define WR_APC2		0x65			// Write D11:D0 into the APC reg, VOL from APC D2:D0

// 4-byte commands
#define RD_PLAY_PTR	0x06			// Returns status bits & current row counter & playback pointer
#define RD_REC_PTR	0x08			// Returns status bits & current row counter & record pointer
#define RD_APC		0x44			// Returns status bits & current row counter & APC register

// 7-byte commands
#define SET_PLAY	0x80			// Play start S10:S0 to end E10:E0 or EOM or STOP
#define SET_REC		0x81			// Record start S10:S0 to end E10:E0 or STOP
#define SET_ERASE	0x82			// Erase from start S10:S0 to end E10:E0

// command modifier bits
#define	LED			0x10			// ISD1700 LED indicator on

// SR0 bits
#define	SR0_CMD_ERR	(1<<0)			// Command error
#define	SR0_FULL	(1<<1)			// Memory full
#define	SR0_PU		(1<<2)			// Device is powered and operating in SPI mode
#define	SR0_EOM		(1<<3)			// this bit is detected when EOM is detected, cleared by CLR_INT command
#define	SR0_INT		(1<<4)			// This bit is set when current operation is done, cleared by CLR_INT command 

// SR1 bits
#define	SR1_RDY		(1<<0)			// Device is ready to accept new command
#define	SR1_ERASE	(1<<1)			// Current operation is erase
#define	SR1_PLAY	(1<<2)			// Current operation is playback
#define	SR1_REC		(1<<3)			// Current operation is recording
#define	SR1_SE1		(1<<4)			// SE1 is recorded
#define	SR1_SE2		(1<<5)			// SE2 is recorded
#define	SR1_SE3		(1<<6)			// SE3 is recorded
#define	SR1_SE4		(1<<7)			// SE4 is recorded

// APC register bits
#define APC_VOL0	(1<<0)			// Volume control, 000 = max
#define APC_VOL1	(1<<1)			// -4dB / bit
#define APC_VOL2	(1<<2)			//
#define APC_MONITOR	(1<<3)			// Enable input signal to output during record
#define APC_MIX_INPUT (1<<4)		// Input Mix control
#define APC_SE_EDITING (1<<5)		// Disable Sound Effect editing
#define APC_SPI_FT	(1<<6)			// nFT bit in SPI mode, 0=feed through
#define APC_AUX		(1<<7)			// AUD/AUX output select, 1 = AUX
// note D11:D8 shifted bit#-8
#define APC_PWM_SPK	(1<<0)			// D8, PWM speaker amp disable, 1 = disable
#define APC_PU_AOUT	(1<<1)			// D9, Analog output disable in PU mode, 1 = disable
#define APC_VALERT	(1<<2)			// D10, Voice allert on/off, 1=off
#define APC_EOM_ENA	(1<<3)			// D11, EOM enable, 1 = EOM enabled


// some APC register bit combination definitions
// APC low data must is shifted << 8 and ored with WR_APC2 command

// APC high bits D11:D8 are same for all commands
// PWM_SPK = 1(disabled), PU_AOUT = 0(enabled), VALERT = 1(off), EOM_ENA = 1
#define APC_HI_init	(APC_PWM_SPK | APC_VALERT | APC_EOM_ENA)


// Initial state / Record from RX / message TX on the air
// Vol = max, Monitor = 1(enabled), Mix = 0, SE_Edit = 0, SPI_FT = 0(on), AUX = 1
#define APC_LO_init	(WR_APC2 | ((APC_MONITOR | APC_AUX) << 8))

// Record from Microphone
// SPI_FT = 1(off) = Mic REC, MONITOR = 1(disabled) 
#define APC_LO_mic (WR_APC2 | ((APC_SPI_FT | APC_AUX) << 8))

// Listen message (Play)
//  SPI_FT = 1(off), MONITOR = 1(disabled)
#define APC_LO_play	(WR_APC2 | ((APC_MONITOR | APC_SPI_FT | APC_AUX) << 8))


// ISD1700 driver variables
int isd_start_addr;					// voice memory start address
int isd_end_addr;					// voice memory end address


// Helper functions
int check_stop(void)				// check if STOP signal 
	{
	int ret;
	ret = 0;

	if(eeprom.defval.serial_mode == 2)	// UART1 mode = ext kb
		{
		if(kbhit())
			ret = 1;				// Stop / new command
		}

	if(kbhit2())					// UART2 allways active
		ret = 1;

	return ret;
	}


// init SPI 2 software driver system
void init_spi2(void)
	{
	ISD_SS = 1;					// set SS high
	ISD_SCLK = 1;				// SCLK idle state = 1
	ISD_MOSI = 0;				// ISD1700 spec !
	}


// exchange SPI 2 data primitive
unsigned int exchg_data_spi2(unsigned int dta)
	{
	int bit_ctr;
	unsigned int rx;

// Exchange data with ISD1700
	for(bit_ctr=0; bit_ctr<16; bit_ctr++)	
		{

		ISD_SCLK = 0;							// generate clock pulse

// set MOSI bit
		if(dta & 0x0001)						// start from bit #0, LSB first
			ISD_MOSI = 1;						// evaluate bit
		else
			ISD_MOSI = 0;
	
		dta = dta >> 1;							// addjust data for next round

// test MISO bit & shir it to rx shift register
		rx = rx >> 1;							// shift old bit in
		if(ISD_MISO)							// get bit from AFEDRI8201 when clk high
			rx = rx | 0x8000;

		ISD_SCLK = 1;		
		}

	return rx;									// return with RX data
	}


// send / receive 16 bit data from SPI 2
unsigned int tx_rx_16_spi2(unsigned int dta)
	{
	unsigned int rx;

	ISD_SS = 0;									// select chip
	rx = exchg_data_spi2(dta);					// do exchange
	ISD_SS = 1;									// deselect chip
	ISD_MOSI = 0;								// ISD1700 spec !
	return rx;
	}

// send command and get responce, 32 bit transfer
unsigned int tx_rx_32_spi2(unsigned int cmd1, unsigned int dta2 )
	{
	unsigned int resp;

	ISD_SS = 0;									// select chip
	exchg_data_spi2(cmd1);						// send command & Data Byte1, discard rx data
	resp = exchg_data_spi2(dta2);				// read responce, TX data = DataByte2
	ISD_SS = 1;									// deselect chip
	ISD_MOSI = 0;								// ISD1700 spec !
	return resp;
	}

// 56bit (7 byte) memory access commands
// resp = SR0 byte1, byte2 
// cmd = 0x0000 + cmd bits
// dta1 = start adress, S10:S0
// dta2 = end aderss, E10:E0

unsigned int tx_rx_56_spi2(unsigned int cmd, unsigned int dta1, unsigned int dta2 )
	{
	unsigned int resp;
	int bit_ctr;

	ISD_SS = 0;									// select chip
	resp = exchg_data_spi2(cmd);				// send command & Data Byte1, discard rx data
	resp = exchg_data_spi2(dta1);				// transfer Start Adres
	resp = exchg_data_spi2(dta2);				// transfer End Adres, read & save responce

		ISD_MOSI = 0;
	for(bit_ctr=0; bit_ctr<8; bit_ctr++)		// send 8 zero bits
		{
		ISD_SCLK = 0;							// generate clock pulse
		__asm__ volatile ("nop");
		__asm__ volatile ("nop");				// 400ns high/low spec
		ISD_SCLK = 1;
		__asm__ volatile ("nop");
		__asm__ volatile ("nop");
		}
	ISD_SS = 1;									// deselect chip
	ISD_MOSI = 0;								// ISD1700 spec !
	return resp;
	}

// Init ISD1700
void init_isd1700(void)							// Init voice memory system
	{
	int devid;

	ISD_RESET = 1;								// remove hardware reset
	init_spi2();
	tx_rx_16_spi2(PU);							// power up device
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 	// init ISD17240
	devid = (0x00FF & tx_rx_32_spi2(DEVID, 0));	// try to read Device ID
	isd_1700_installed = 0;
	if(devid == ISD_17240)
		{
		tx_rx_32_spi2(APC_LO_init, APC_HI_init); 	// init ISD17240
		printf("\n\rTRX-2 voice memory option found (ISD17240 chip)\n\r");		
		isd_1700_installed = 1;
		}
	}


// Set Start & End address
void set_isd_addresses(int idx)
	{
	isd_start_addr = ISD_MEM [idx] [0];											// convert index to physical addresses
	isd_end_addr = ISD_MEM [idx] [1];

//	printf("\n\rStar = %X, End = %X\n\r",isd_start_addr, isd_end_addr);			// debug printouts
//	printf("\n\rStar = %i, End = %i\n\r",isd_start_addr, isd_end_addr);			// print start & stop address
	}

// Play selected memory block
// mode 0=play, 1=TX
void play_mem(int idx, int mode)
	{
	seconds = 0;																// reset playback timer
	clear_lcd();
	if(mode == 0)
		{
		sprintf(lcdpbuff, "PLAY Mem # %i  ", idx);								// Display function & mem # 
		}
	else
		{
		start_scaf_clock();														// audio on, if squelch closed
		sq_timer = SQ_HOLD;														// reload squelch open timer
		sprintf(lcdpbuff, "TX M %i  ", idx);									// Display function & mem # 
		}

	lcd_putst(lcdpbuff);
	tx_rx_32_spi2(APC_LO_play, APC_HI_init); 									// RX mute, audio from ISD1700

	set_isd_addresses(idx);														// set ISD1700 addresses
	tx_rx_16_spi2(0x0000 | CLR_INT);											// clear INT signal
	tx_rx_56_spi2((0x0000 | SET_PLAY | LED),isd_start_addr , isd_end_addr );	// execute command

	while(ISD_RDY != 0)															// wait for Ready signal
		{
		if(mode != 0)															// if TX
			{
			set_cur_lcd(LINE1+8);												// measure & display power meter
			fwd_adval = convert_adc12(FWD_PWR);
			disp_bar_meter();
			}
		set_cur_lcd(LINE2);
		sprintf(lcdpbuff, "Time %i s      ", seconds);							// Display elapsed time
		lcd_putst(lcdpbuff);
		if(check_stop() != 0)													// stop pressed
			{
			tx_rx_16_spi2(STOP | LED);											// Stop command
			break;																// leave on any key
			}
		}
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 									// audio in = Receiver audio
	}


// Erase & Record selected memory block
// ch0 = mic, ch1=receiver
void rec_mem(int idx, int ch)
	{
	seconds = 0;																// reset record timer
	clear_lcd();
	if(ch == 0)																	// Record from Microphone
		{
		sprintf(lcdpbuff, "REC Mic  Mem # %i ", idx);							// Display function & mem #
		tx_rx_32_spi2(APC_LO_mic, APC_HI_init); 								// audio in = Microphone
		}
	else																		// record from RX
		{
		start_scaf_clock();														// audio on, if squelch closed
		sq_timer = SQ_HOLD;														// reload squelch open timer
		sprintf(lcdpbuff, "RX M %i  ", idx);									// Display function & mem #
		tx_rx_32_spi2(APC_LO_init, APC_HI_init); 								// audio in = Receiver audio
		}
	lcd_putst(lcdpbuff);
	set_cur_lcd(LINE2);

	set_isd_addresses(idx);														// set ISD1700 addresses
	tx_rx_16_spi2(0x0000 | CLR_INT);											// clear INT signal
	tx_rx_56_spi2((0x0000 | SET_ERASE | LED),isd_start_addr , isd_end_addr );	// erase memory range

	while(ISD_RDY != 0)															// wait for INT signal
		{
		}
	tx_rx_16_spi2(0x0000 | CLR_INT);											// clear INT signal

	tx_rx_56_spi2((0x0000 | SET_REC | LED),isd_start_addr , isd_end_addr );		// Start recording

	while(ISD_RDY != 0)															// wait for INT signal
		{
		if(ch != 0)																// if RX
			{
			set_cur_lcd(LINE1+8);												// measure & display S-meter
			disp_bar_meter();
			}
		set_cur_lcd(LINE2);
		sprintf(lcdpbuff, "Time %i s      ", seconds);							// Display elapsed time
		lcd_putst(lcdpbuff);
		if(check_stop() != 0)													// stop pressed
			{
			tx_rx_16_spi2(STOP | LED);											// Stop command, write EOM
			break;																// leave here
			}
		}
	tx_rx_16_spi2(CLR_INT);														// clear INT signal
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 									// audio in = Receiver audio
	}
