//
// dsPIC30 timers PWM DAC for JUMA-TRX2
// Juha Niinikoski, OH2NLT 30.11.2006
//
// Microchip C30 compiler conversion
//
// JUMA-TRX2 timer & pwm system
//
// Timer & PWM unit usage
// TMR1 is used for tone generation, IRQ driven system, RD1/OC2 = tone out
// TMR2 is uded for SCAF clock generation
// OC1 output is for SCAF filter clock
// OC1 is connected to TMR2
// TMR3 is used for 1ms timer tick IRQ and for LCD PWM dac timebase
// OC3 output is for LCD back light control
// OC4 output is for LCD contrast
// OC3 and OC4 are connected to TMR3
//
// TMR4 reserved for CW timing
// TMR5 used for delays (separate module)
// Click less tone generator code 16.12.2006
// Tune mode start clitch removed 15.01.2007
// 1s counter added for ISD1700 routines
// tone generator end click problem corrected 17.03.2007
// Yaesu CAT protocol timeout counter 22.01.2008
// addjustable CW sidetone 02.01.2008
// LCD back light timer aded 03.07.2008
// __attribute__((auto_psv)) added to the IRQ functions 05.03.2015

#define TICK_PERIOD 7500		// 1ms tick

#define SCAF_05 150				// 500Hz / 50kHz clock
#define SCAF_19 39				// 1,9kHz / 192,307kHz clock
#define SCAF_24 31				// 2,4kHz / 241,9kHz clock
#define SCAF_25 30				// 2,5kHz / 250kHz clock


#include "juma-trx2.h"

static unsigned int ms_counter;		// general mod 65536 ms counter
static unsigned int tone_counter; 	// beep tone length (ms)
static unsigned int busy_counter;	// tone generator busy(ms), busy if != 0
static int key_down;				// key down flag for straight mode tone generator
static int last_cycle;				// for click less end trick

static int iambic_flag;				// last played, 0=dot, 1=dash
static int dot;						// DOT RS flip flop, follow paddle logic, 0=pressed
static int dash;					// DASH RS flip flop

int cw_break_timer;					// cw break timer
int cw_period;						// CW period(ms)

int cmd_timeout;					// RS232 command timeout, 0 = timeout, negative values = state flags.
unsigned int pa100_tx_tmr;			// TRX2 protocol band data TX rate

unsigned int long_push;				// long push timer, count from set value to zero, global visibility

// external variables

extern int cw_active;				// cw active flag for keyer code
extern int keyer_mode;				// 0=dot priority, 1=iambic A,  2=iambic B, 3=straight

extern int seconds;					// free running second counter

extern int sq_timer;				// squelch timer
extern int bl_timer;				// LCD back light timer
extern int bl_off_flag;				// turn backlight off flag


// EEPROM structures

#include "trx2_eeprom.h"			// get EEPROM structure definitions

extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval)/2];
	}eeprom;


// init timer & PWM system
void init_timers_pwm(void)
	{
	dot = 1;					// reset paddle flip flops
	dash = 1;

	ms_counter = 1000;
	OC3CON = 0;					// first turn off the module if it was on
	OC4CON = 0;

// LCD PWM DAC setup
	OC3RS = 100;				// some small value for back light
	OC3R = 100;
	OC3CON = 0x000E;			// use TMR3, OC3 = PWM mode

	OC4RS = 2000;				// something for contrast to make display visible
	OC4R = 2000;
	OC4CON = 0x000E;			// use TMR3, OC4 = PWM mode

	PR3 = TICK_PERIOD - 1;		// set PWM cycle, TMR3 period register, 1000Hz/1ms
	T3CONbits.TON = 1;			// start TMR3

	_T3IE = 1;					// enable TMR3 tick irq
	}

void init_scaf_clock(void)
	{
// SCAF clock
	OC1CON = 0;					// first turn off the module if it was on
	OC1RS = SCAF_25/2;			// set OC1 duty cycle to 50%
	OC1R = SCAF_25/2;
	OC1CON = 0x0006;			// use TMR2, OC1 = PWM mode
	PR2 = SCAF_25;				// SCAF clock frequency, wide filter
	T2CONbits.TON = 1;			// start TMR2	
	}

void stop_scaf_clock(void)
	{
	T2CONbits.TON = 0;			// stop TMR2
	}

void start_scaf_clock(void)
	{
	T2CONbits.TON = 1;			// start TMR2
	}


unsigned int get_ms_counter(void)
	{
	return ms_counter;
	}


// PWM set functions
void set_scaf_clock(unsigned int scaf)	// SCAF clock
	{
	if(scaf < 2)
		scaf = 2;						// protect PR2 setup
	OC1RS = scaf >> 1;					// try to keep 50% duty cycle
	PR2 = scaf-1;
	}

void set_pwm3_dac(unsigned int pwm)		// LCD back light
	{
	OC3RS = pwm;
	}

void set_pwm4_dac(unsigned int pwm) 	// LCD contrast
	{
	OC4RS = pwm;
	}

// Tone generator
void __attribute__((interrupt)) __attribute__((auto_psv)) _T1Interrupt(void)
	{
//	_T1IF = 0;

	switch(last_cycle)					// tone generator state machine
		{
		case 0: default:				// tone output off, tri state
		TONE_TRIS = 1;					// tristate output
		_T1IE = 0;						// disable TMR1 IRQs, Tone generator off
		break;

		case 1:							// last half cycle
		TONE_OUT = !TONE_OUT;
		PR1 = PR1 >> 1;					// next = half cycle
		last_cycle = 0;					// flag for  shut down
		break;

		case 2:							// last full cycle, start shut down
		TONE_OUT = !TONE_OUT;			// inverse tone out on every cycle
		if(TONE_OUT != 0)				// end with down going half cycle
			last_cycle = 1;				// next = last full cycle
		break;

		case 3:							// generate sound
		TONE_OUT = !TONE_OUT;			// inverse tone out on every cycle
		break;

		}

	_T1IF = 0;

	}

// Tone set functions
void tone_on(int tone)					// set tone, set value = 7,5MHz / tone(Hz) * 2
	{
	last_cycle = 3;						// not last cycle, run state
	PR1 = tone - 1;						// set tone period
	TMR1 = PR1 >> 1;					// star with half period, reduce klick
	TONE_OUT = 0;						// start allways same phase
	TONE_TRIS = 0;						// enable output
	T1CONbits.TON = 1;					// start timer
	_T1IF = 0;							// clear old pending interrupt, star with clean half cycle
	_T1IE = 1;							// enable TMR1 IRQs, Tone generator on
	}

void tone_off(void)						// stop tone generator
	{
	last_cycle = 2;						// set flag forstop tone, last half cycle
	}


void beep(int tone, int duration)
	{
	if(busy_counter == 0 && duration > 0 && key_down == 0)	// start tone only if tone generator free & duration specified
		{
		start_scaf_clock();				// open audio path
		sq_timer = SQ_HOLD;				// keep squelch open
		tone_counter = duration;		// set tone length
		tone_on(tone);					// set tone pitch & start play
		}
	}


// Keyer code

void play_dot(void)						// Turn TX on & generate dot
	{
	iambic_flag = 0;					// remember last played
	if(cw_active != 0)
		{
		PTT_OUT = 1;					// RF output on if cw mode
		cw_break_timer = 7 * cw_period;	// TX on break time
		KEY = 1;						// CW modulator on
		}
	beep(eeprom.defval.cw_sidetone, cw_period);			// start tone
	busy_counter = 2 * cw_period;		// set busy period
	}

void play_dash(void)
	{
	iambic_flag = 1;					// remember last played
	if(cw_active != 0)
		{
		PTT_OUT = 1;					// RF output on if cw mode
		cw_break_timer = 7 * cw_period;	// TX on break time
		KEY = 1;						// CW modulator on
		}
	beep(eeprom.defval.cw_sidetone, (3 * cw_period));	// start tone
	busy_counter = 4 * cw_period;		// set busy period
	}

// IRQ code
// 1ms Tick timer
void __attribute__((interrupt)) __attribute__((auto_psv)) _T3Interrupt(void)
	{
	_T3IF = 0;

	if(tone_counter != 0)
		{
		tone_counter--;
		if(tone_counter == 0)
			{
			tone_off();					// run stop with half cycle logic
//			TONE_TRIS = 1;				// tristate output
//			_T1IE = 0;					// disable TMR1 IRQs, Tone generator off
			if(eeprom.defval.mode[eeprom.defval.vfo_sel] != 3)	// if tune MODE selected, do not touch KEY bit
				KEY = 0;				// turn CW modulator off if any other MODE
			}
		}

	if(cw_break_timer != 0)				// run cw break timer
		{
		cw_break_timer--;
		if(cw_break_timer == 0)
			{
//			if(cw_active != 0)			// condition removed 03.07.2008
				PTT_OUT = 0;			// RF output off if cw mode selected
			}
		}

	if(busy_counter != 0)				// decrement CW tone busy timer if active
		{
		busy_counter--;
		
		if(busy_counter == 0)			// mark ready
			{
			if(iambic_flag == 0)		// was DOT
				dot = 1;				// reset dot
			else						// was dash
				dash = 1;				// reset dash
			}
		}

	if(long_push != 0)					// decrement long push timer if active
		long_push--;

// read paddle switches, set dot/dash RS flip flops
	if(keyer_mode == 2)					// Iambic B, flip set only
		{
		if(DOT == 0)					// Set RS/ flip flop
			dot = 0;

		if(DASH == 0)
			dash = 0;
		}
	else								// Iambic A or other mode, flip flop set/reset
		{
		if(DOT == 0)					// Set/Reset RS flip flop
			dot = 0;
		else
			dot = 1;

		if(DASH == 0)
			dash = 0;
		else
			dash = 1;
		}

// run keyer logic
		switch(keyer_mode)				// select keyer operation
			{
			case 0: default:			// dot priority
			if(DOT == 0 && busy_counter == 0)	// start only if not sound going
				play_dot();
			else
				{
				if(DASH == 0 && busy_counter == 0)
					play_dash();
				}
			break;


			case 1:	case 2:							// iambic mode A & B, sample from dot/dsah RS flip flop
			if(iambic_flag != 0)					// last played was dash, test DOT contact first
				{
				if(dot == 0 && busy_counter == 0)	// start only if not sound going
					play_dot();
				else
					{
					if(dash == 0 && busy_counter == 0)
						play_dash();
					}
				}
			else									// last played was dot, test DASH contact first
				{
				if(dash == 0 && busy_counter == 0)	// start only if not sound going
					play_dash();
				else
					{
					if(dot == 0 && busy_counter == 0)
						play_dot();
					}
				}
			break;


			case 3:									// straight mode
			if(DOT == 0)							// key down, I/O bit
				{
				if(key_down == 0)					// do this only if key was up
					{
					key_down = 1;					// set key down status
					tone_on(eeprom.defval.cw_sidetone);				// keep on starting side tone generator
					}

				if(cw_active != 0)					// tx only if CW mode selected
					{
					PTT_OUT = 1;					// RF output on if cw mode
					cw_break_timer = 7 * cw_period;	// TX on break time
					KEY = 1;						// CW modulator on
					}
				}

			if(DOT != 0)							// key up
				{
				if(key_down != 0)					// do only if key was down
					{
					KEY = 0;						// CW modulator off
					tone_off();						// stop side tone generator
					key_down = 0;					// remove key down flag
					}
				}
			break;
			}

// general timing
	ms_counter--;									// run 1s free running counter
	if(ms_counter == 0)
		{
		ms_counter = 1000;
		seconds++;
		if(bl_timer != 0)							// run LCD back light timer
			{
			bl_timer--;
			if(bl_timer == 0)
				bl_off_flag = 1;
			}
		}

// squelch timer
	if(sq_timer != 0)								// run squelch timer
		sq_timer--;


// RS232 timeout timer for Yaesu CAT protocol	
	if(cmd_timeout > 0)								// negative values used as state flags
		cmd_timeout--;

// TRX2 protocoll timer
	if(pa100_tx_tmr != 0)							// used to limit band info TX rate to the PA100
		pa100_tx_tmr--;
	}
