//
// Simple UART interface for dsPIC30F6012
// Juha Niinikoski, OH2NLT 18.05.2005
//
// Below is the code to initialize the USART
// 8 bits, 1 start and 1 stop, no error handling
// kbhit and getch added 22.05.2005
// getche aded 02.06.2005unsigned char
// overrun error handling added 29.08.2005
// MPLAB C30 conversion, OH2GWE 2006.10.17
// Baud Rate setup & IRQ receiver for fast baud rates 14.01.2008
// Compiler bug work arround 01.02.2008
// UART1 init sequence changed to be compatible with later silicon 30.10.2008
// data type cast corrections 18.11.2008
// __attribute__((auto_psv)) added to the IRQ 05.03.2015

#include "juma-trx2.h"
#include <stdio.h>

//#define BAUD 9600
//#define FCY  1000000	// 4MHz XTAL
//#define FCY  8000000	// 4MHz XTAL x8 PLL = 32MHz
//#define FCY  16000000	// 4MHz XTAL x16 PLL = 64MHz
//#define FCY  29491200	// 7,3728MHz XTAL x16 PLL = 117,9648MHz


// Baud Rate divisors for 30 MHz osc =  7,5 MHz FCY 
/* actual baud rate divisors
[0] 1200	389,625
[1] 2400	194,3125
[2] 4800	96,65625
[3] 9600	47,828125
[4] 19200	23,4140625
[5] 38400	11,20703125
[6] 57600	7,138020833
[7] 115200	3,069010417
*/

/* Module variables & definitions */

static volatile unsigned char rx_buffer[256];      // UART mod(256) RX queue, must be exactly 256 long
static volatile unsigned char rx_buf_out_idx = 0;  // buffer output index, pointer behind in idx
static volatile unsigned char rx_buf_in_idx = 1;   // buffer input index, pointer always "leading"

const unsigned int baud_rates[8] = {390, 194, 97, 48, 23, 11, 7, 3};



// UART1 IRQ service
// queue character
void __attribute__((interrupt)) __attribute__((auto_psv)) _U1RXInterrupt(void)  // put receiced character to RX queue
	{
  	unsigned char rx;

  	rx = U1RXREG;                        			// data read
  	if((unsigned char)(rx_buf_in_idx + 1) != rx_buf_out_idx) 		// check overrun
  		{
 		rx_buffer[rx_buf_in_idx] = rx;    		   // place just received character in queue
		rx_buf_in_idx++;
    	}
  	// else reject character

// check & handle possible errors
	if(U1STAbits.OERR == 1)							// OERR is blocking uart, should not happen but..
		U1STAbits.OERR = 0;							// reset overrun error if occured

// this is not automatic in 30F6014 UART
	IFS0bits.U1RXIF = 0;							// reset irq flag
	}


// UART1 init
void InitUSART1(void)
{
 //U1BRG = ((FCY/16)/BAUD) - 1;   		// set baud rate
 U1BRG = baud_rates[3];					// set intial baud rate to 9600

 U1MODE = 0x0000;
 U1STA = 0x0000;						// IRQ on every character

// U1STAbits.UTXEN = 1;					// enable transmitter
 IFS0bits.U1RXIF = 0;					// clear possible pending irq
 IEC0bits.U1RXIE = 1;					// enable RX IRQs
 U1MODEbits.UARTEN = 1;					// enable uart
 U1STAbits.UTXEN = 1;					// enable transmitter, must be after UARTEN for later silicon
}

// set baud rate
void SetUSART1baud(int baud_sel)
	{
	if(baud_sel > 7)
		baud_sel = 7;					// stay in the table

	 U1BRG = baud_rates[baud_sel];		// set baud rate
	}

// print character
void putch(char c)
	{
	while(U1STAbits.UTXBF == 1)		// wait here if tx fifo full
		{
		}
	U1TXREG = (int)c &0x00FF;	// Tx reg is 16-bit long
	}

#if 0
// check if key perssed, data available from UART receive buffer
// Not pressed = 0, Pressed = 1
// non IRQ function
int kbhit(void)
	{
	if(U1STAbits.URXDA)
		return 1;
	else
		return 0;
	}
#endif

// check if key perssed, data available from receive buffer
// Not pressed = 0, Pressed = number of characters
// IRQ based RX queue check
int kbhit(void)
	{
	unsigned char ret;

//	if(U1STAbits.OERR == 1)								// OERR is blocking uart, should not happen but..
//		U1STAbits.OERR = 0;								// reset overrun error if occured
//  	return (rx_buf_in_idx  - (rx_buf_out_idx + 1));   	// return with number of available characters, compiler does error here !

	ret = rx_buf_in_idx  - (unsigned char)(rx_buf_out_idx + 1); // return with number of available characters
	return(ret);
	}

// clear serial port, reset input buffer
void ClearUSART1queue(void)
	{
	rx_buf_out_idx = 0;  								// buffer output index
	rx_buf_in_idx = 1;   								// buffer input index
	}


#if 0
// receive character, non IRQ function
unsigned char getch(void)
	{
	while(U1STAbits.URXDA == 0)		// wait for character
		{
		}
// printf("U1STAbits = 0x%x\n\r", U1STAbits); // debug
	IFS0bits.U1RXIF = 0;	// reset irq flag
	if(U1STAbits.OERR == 1)		// OERR is blocking uart
		U1STAbits.OERR = 0;		// reset overrun error if occured
	return U1RXREG;			// return with character
	}
#endif

// get char from RX queue, IRQ RX system
unsigned char getch(void)
	{
 	while((unsigned char)(rx_buf_out_idx + 1) == rx_buf_in_idx);   	// wait data if queue is empty 
  	return rx_buffer[++rx_buf_out_idx];             	// return with data
	}


// receive character with echo
unsigned char getche(void)
	{
	unsigned char c;
	putch(c = getch());
	return c;
	}



// some debug I/O functions
// Get unsigned long

unsigned long getlong(void)
	{
	unsigned long val;
	char tmp;
	val = 0;
	
	do
		{
		tmp = getch() & 0x7F;
		putch(tmp);
		if(tmp >= '0' && tmp <= '9')
			val = 10 * val + (tmp - '0');
		}
	while (tmp != 0x0D);
	
	return(val);
	}

// Special output functions

unsigned char disp_buffer[11];		// display buffer for long number printouts

/* Convert binary number to ASCII string */
/* 10 digits are converted */

void bin2ascii(unsigned long luku)
	{
char x, idx;

	idx = 9;
	for(x=0; x<10; x++)				// do 10 digits
		{
		disp_buffer[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	disp_buffer[10] = 0;			// string terminator
	}

void putlong( unsigned long x)
	{
	bin2ascii(x);					// partial convert long number to string

	putch(disp_buffer[0]);		// MSB

	putch(disp_buffer[1]);		// 
	putch(disp_buffer[2]);		// 
	putch(disp_buffer[3]);		//

	putch(disp_buffer[4]);		//
	putch(disp_buffer[5]);		//
	putch(disp_buffer[6]);		//

	putch(disp_buffer[7]);		//
	putch(disp_buffer[8]);		//
	putch(disp_buffer[9]);		// LSB
	}


// Uart/rs232 input debug
void rs232_debug(void)
	{
	int x;
	
	printf("\n\rOut_ixd %i, In_idx %i  \n\r",rx_buf_out_idx, rx_buf_in_idx);

	for(x=0; x<255; x++)
		{
		putch(rx_buffer[x]);
		}
	}
